unit addon;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Spin, DateUtils, Types;

{$SetPEFlags IMAGE_FILE_RELOCS_STRIPPED}
{$WEAKLINKRTTI ON}

type
  TTimeType = (ttCreation, ttLastAccess, ttLastWrite);

type
  TForm1 = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    SecondsSpinEdit: TSpinEdit;
    YearsSpinEdit: TSpinEdit;
    MonthsSpinEdit: TSpinEdit;
    DaysSpinEdit: TSpinEdit;
    HoursSpinEdit: TSpinEdit;
    MinutesSpinEdit: TSpinEdit;
    StartButton: TButton;
    CancelButton: TButton;
    cbLastWrite: TCheckBox;
    cbCreation: TCheckBox;
    cbLastAccess: TCheckBox;
    cbSetDirectoriesDateTime: TCheckBox;
    cbRecurseSubdirectories: TCheckBox;
    procedure CancelButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    procedure FormKeyPress(Sender: TObject; var Key: Char);
    procedure cbLastAccessClick(Sender: TObject);
  private
    function MsgBox(const Text: string; uType: Cardinal = MB_OK): Integer;
    procedure ProcessDirEx(const StartPath: string);
    function SetFileDateTime(const FileName: string; FileHandle: THandle;
      TimeType: TTimeType): Boolean;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
  filelist: tstringlist;

implementation

{$R *.dfm}

uses IOUtils;

function GetVeryLongFileName(const FullFileName: string): string; inline;
begin
  Result := FullFileName;
  if Length(FullFileName) < 256 then Exit;
  if AnsiPos('\\?\', FullFileName) > 0 then Exit;

  if AnsiPos('\\', FullFileName) > 0 then
    Result := '\\?\UNC\' + Copy(FullFileName, 3, Length(FullFileName) - 2)
  else
    Result := '\\?\' + FullFileName;
end;

procedure ReadFileNames;
var
  fileinfo: tsearchrec;

begin

  // Let's add all the files we can find in the current dir

  if findfirst('*.*', $27, fileinfo) = 0 then
    repeat
      filelist.Add(expandfilename(fileinfo.name));
    until findnext(fileinfo) <> 0;
  findclose(fileinfo);
end;

procedure ProcessDir(startpath: string);
var
  fileinfo: tsearchrec;                         // Used in findfirst / findnext
  i: integer;                                   // Used to check the result of findnext
  curdir: string;                   // To save the name of the dir we're in when going up '..'

begin
  ReadFileNames;                                // Adds all the files in the current dir


  // let's find the first subdir

  findfirst('*.*', $10, fileinfo);
  findnext(fileinfo);
  repeat
    i := findnext(fileinfo);
  until (i <> 0) or (fileinfo.attr and faDirectory <> 0);
  findclose(fileinfo);


  // i determines if subdir was found. 0 means subdir was found, any other value
  // means no subdir found

  // let's keep going deeper and deeper into a subdir and processing each subdir
  // until no further subdir is found (subdir subdir subdir subdir!)

  if i = 0 then
    begin
      chdir(fileinfo.name);
      processdir(startpath);
    end


  // if no deeper subdir is found then let's find out the name of the current dir,
  // go one level higher and see if there are more dirs than the one we were in.
  // If yes, then let's go into them and process them
  // We repeat this until we're not back up in the root (start) path

  else
    if getcurrentdir <> startpath then          // Only do this if we aren't in the startpath
      begin
        repeat
          curdir := copy(getcurrentdir, lastdelimiter('\', getcurrentdir) + 1,
           length(getcurrentdir) - lastdelimiter('\', getcurrentdir));
          chdir('..');
          findfirst('*.*', $10, fileinfo);
          findnext(fileinfo);
          repeat
            findnext(fileinfo);
          until (fileinfo.attr and faDirectory <> 0) and (fileinfo.Name = curdir);
          repeat
            i := findnext(fileinfo);
          until (i <> 0) or (fileinfo.attr and faDirectory <> 0);
        until (i = 0) or (getcurrentdir = startpath);
        findclose(fileinfo);

        if i = 0 then                                     // when subdir is found let's
          begin                                           // process it, otherwise end it
            chdir(fileinfo.name);                         // (and possibly return to where
            processdir(startpath);                        // we were called recursively)
          end;
      end;
end;

procedure TForm1.CancelButtonClick(Sender: TObject);
begin
  Application.Terminate;
end;

procedure TForm1.FormKeyPress(Sender: TObject; var Key: Char);
begin
  if Ord(Key) = VK_ESCAPE then Close;                       // Escape pressed
  if Ord(Key) = VK_RETURN then
    if StartButton.Enabled then StartButton.Click;          // Enter pressed
end;

function TForm1.MsgBox(const Text: string; uType: Cardinal = MB_OK): Integer;
begin
  Result := MessageBox(Handle, PChar(Text), PChar(Application.Title), uType);
end;

procedure TForm1.ProcessDirEx(const StartPath: string);
var
  Files: TStringDynArray;
  Str: string;
begin
  if cbSetDirectoriesDateTime.Checked then
    Files := TDirectory.GetFileSystemEntries(StartPath,
      TSearchOption.soAllDirectories, nil)
  else
    Files := TDirectory.GetFiles(StartPath, '*',
      TSearchOption.soAllDirectories);
  for Str in Files do
    filelist.Add(Str);
end;

procedure TForm1.cbLastAccessClick(Sender: TObject);
begin
  StartButton.Enabled := cbLastWrite.Checked or cbLastAccess.Checked or
    cbCreation.Checked;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  {$IFDEF WIN64}
  Application.Title := Application.Title + ' (x64)';
  {$ENDIF}
  Caption := Application.Title;

  if paramcount <> 1 then                       // Check if there is exactly one parameter -
    begin                                       // the name of the file that contains a list
      MsgBox('Paramcount <> 1. Exiting.');      // of all the files to be changed
      halt;
    end;
  if fileexists(paramstr(1)) = false then       // Check if the list file exists
    begin
      MsgBox('List file ''' + paramstr(1) + ''' not found. Exiting.');
      halt;
    end;
end;

function TForm1.SetFileDateTime(const FileName: string; FileHandle: THandle;
  TimeType: TTimeType): Boolean;
var
  date: tdatetime; // Used to store / change the filedates
  lt, tDT: filetime;
  systime: tsystemtime;
  GFTResult: Boolean;
begin
  Result := false;

  case TimeType of
    ttLastAccess: GFTResult := GetFileTime(FileHandle, nil, @tDT, nil);
    ttCreation: GFTResult := GetFileTime(FileHandle, @tDT, nil, nil);
    ttLastWrite: GFTResult := GetFileTime(FileHandle, nil, nil, @tDT);
  else
    GFTResult := GetFileTime(FileHandle, nil, nil, @tDT);
  end;

  if not GFTResult then
  begin
    MsgBox('GetFileTime error on file ' + FileName + #13#13 +
      SysErrorMessage(GetLastError));
    fileclose(FileHandle);
    Exit;
  end;

  if FileTimeToLocalFileTime(tDT, lt) = false then
  begin
    MsgBox('FileTimeToLocalFileTime error on file ' + FileName + #13#13 +
      SysErrorMessage(GetLastError));
    fileclose(FileHandle);
    Exit;
  end;

  if FileTimeToSystemTime(lt, systime) = false then
  begin
    MsgBox('LocalFileTimeToSystemTime error on file ' + FileName + #13#13
      + SysErrorMessage(GetLastError));
    fileclose(FileHandle);
    Exit;
  end;

  if TrySystemTimeToDateTime(systime, date) = false then
  begin
    MsgBox('SystemTimeToDateTime error on file ' + FileName);
    fileclose(FileHandle);
    Exit;
  end;

  date := incyear(date, YearsSpinEdit.Value);
  date := incmonth(date, MonthsSpinEdit.Value);
  date := incday(date, DaysSpinEdit.Value);
  date := inchour(date, HoursSpinEdit.Value);
  date := incminute(date, MinutesSpinEdit.Value);
  date := incsecond(date, SecondsSpinEdit.Value);

  datetimetosystemtime(date, systime);

  if systemtimetofiletime(systime, lt) = false then
  begin
    MsgBox('SystemTimetoLocalFileTime error on file ' + FileName + #13#13
      + SysErrorMessage(GetLastError));
    fileclose(FileHandle);
    Exit;
  end;

  if LocalFileTimeToFileTime(lt, tDT) = false then
  begin
    MsgBox('LocalFileTimeToFileTime error on file ' + FileName + #13#13 +
      SysErrorMessage(GetLastError));
    fileclose(FileHandle);
    Exit;
  end;

  case TimeType of
    ttLastAccess: GFTResult := SetFileTime(FileHandle, nil, @tDT, nil);
    ttCreation: GFTResult := SetFileTime(FileHandle, @tDT, nil, nil);
    ttLastWrite: GFTResult := SetFileTime(FileHandle, nil, nil, @tDT);
  else
    GFTResult := SetFileTime(FileHandle, nil, nil, @tDT);
  end;

  if not GFTResult then
  begin
    MsgBox('Error setting time of ' + FileName + #13#13 +
      SysErrorMessage(GetLastError));
    fileclose(FileHandle);
    Exit;
  end;

  Result := True;
end;

procedure TForm1.StartButtonClick(Sender: TObject);
var
  f: TStreamReader; // The list file
  s: string;        // Used in readln
  i: cardinal;      // Used when iterating over the filelist
  handle: THandle;
  tLastWrite, tLastAccess, tCreation: filetime;
begin
  if (YearsSpinEdit.Value = 0) and (MonthsSpinEdit.Value = 0) and
    (DaysSpinEdit.Value = 0) and (HoursSpinEdit.Value = 0) and
    (MinutesSpinEdit.Value = 0) and (SecondsSpinEdit.Value = 0) then
  begin
    if MsgBox('All values is 0. Nothing change. Continue?',
      MB_YESNOCANCEL or MB_DEFBUTTON1 or MB_ICONQUESTION) <> ID_YES then
      Exit;
  end;

  filelist := TStringList.Create;
  try
    f := TStreamReader.Create(paramstr(1), TEncoding.ANSI, True);
    try
      filelist.CaseSensitive := false;
      filelist.Capacity := 100;
      filelist.Sorted := True;          // Must be true so that the next line works
      filelist.Duplicates := dupIgnore; // Make sure we don't process a file twice


      // Let's read all the filenames into the filelist

      while not f.EndOfStream do
      begin
        s := f.ReadLine;
        if s[length(s)] <> '\' then // Is a file
        begin
          if not AnsiSameText(s,paramstr(0)) then  // Make sure we don't add ourselves
            filelist.Add(GetVeryLongFileName(s));
        end
        else // Is a dir
        begin
          //chdir(s);
          //processdir(getcurrentdir);
          if cbSetDirectoriesDateTime.Checked then
            filelist.Add(GetVeryLongFileName(s));
          if cbRecurseSubdirectories.Checked then
            ProcessDirEx(GetVeryLongFileName(s));
        end;
      end;

    finally
      f.Free;
    end;

    if filelist.Count > 0 then
      for i := 0 to filelist.Count - 1 do
      begin
        // date := filedatetodatetime(fileage(filelist[i]));

        if cbSetDirectoriesDateTime.Checked and DirectoryExists(filelist[i]) then
          Handle := CreateFile(PChar(filelist[i]),GENERIC_WRITE,FILE_SHARE_READ,nil,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0)
        else
          Handle := fileopen(filelist[i], fmOpenWrite + fmShareDenyNone);
        if Handle = INVALID_HANDLE_VALUE then
        begin
          MsgBox('Error opening ' + filelist[i] + #13#13 +
            SysErrorMessage(GetLastError));
          continue;
        end;

        if cbLastWrite.Checked then
          if not SetFileDateTime(filelist[i], Handle, ttLastWrite) then
            Continue;
        if cbCreation.Checked then
          if not SetFileDateTime(filelist[i], Handle, ttCreation) then
            Continue;
        if cbLastAccess.Checked then
          if not SetFileDateTime(filelist[i], Handle, ttLastAccess) then
            Continue;

        fileclose(Handle);

        // filesetdate(filelist[i], datetimetofiledate(date));
      end;

  finally
    filelist.Free;
  end;

  application.terminate;
end;

end.
